/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "processor.h"

/* {...COMPLEX queueing routines */

extern struct processor *PROC;
#define  MAX(a, b)      ((a) > (b) ? (a) : (b))
#define  MIN(a, b)      ((a) < (b) ? (a) : (b))

int have_fmadd;
int fp_latency;
struct rotating_reg *CMADregs=NULL;

void setup_cmadds(void)
{
  int i;
  char *str;
  fp_latency = 0;
  have_fmadd = 0;
  test_proc();

  for ( i =0 ;i < PROC->npipes; i++ ){
    if ( PROC->pipes[i].issuegroups & FMACMASK ) have_fmadd = 1;
    fp_latency = PROC->latency[FMULPIPE];
  } 
  
  if ( !have_fmadd ) {
    CMADregs = create_rotating_reg(Fregs,MAX ( 1+fp_latency,2),"CMADD");
  }
  return;
}

/*Queues (e + if) = (a+ib)*(c+id)*/
void queue_cmul(int e,int f,int a,int b,int c,int d)
{ 
  if ( have_fmadd ){
    make_inst(FMULPIPE,FMUL,e,a,c);
    make_inst(FMULPIPE,FMUL,f,a,d);
    make_inst(FMACPIPE,FNMSUB,e,b,d,e);
    make_inst(FMACPIPE,FMADD,f,b,c,f);
  }
  else{
    int t;

    make_inst(FMULPIPE,FMUL,e,a,c);

    make_inst(FMULPIPE,FMUL,f,a,d);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,f,f,t);
  }
  return;
}
/*Queues (e + if) = (e+if) + (a+ib)*(c+id)*/
void queue_cmadd(int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,e,a,c,e);
    make_inst(FMACPIPE,FMADD,f,a,d,f);
    make_inst(FMACPIPE,FNMSUB,e,b,d,e);
    make_inst(FMACPIPE,FMADD,f,b,c,f);
  } 
  else {
    int t;

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,f,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,f,f,t);

  }
}


/*Queues (h + ig) = (e+if) + (a+ib)*(c+id)*/
void queue_cmaddto(int h,int g,int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,h,a,c,e);
    make_inst(FMACPIPE,FMADD,g,a,d,f);
    make_inst(FMACPIPE,FNMSUB,h,b,d,h);
    make_inst(FMACPIPE,FMADD,g,b,c,g);
  } 
  else {
    int t;
    /*Want to leave e and f undisturbed so do the copy in the first one*/
    /*Otherwise risk different behaviour on FMADD and non-FMADD chips*/
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,h,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,h,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,h,h,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,g,g,t);

  }
}

/*Queues (e + if) = (a-ib)*(c+id)*/
void queue_conjmul(int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMULPIPE,FMUL,e,a,c);
    make_inst(FMULPIPE,FMUL,f,a,d);
    make_inst(FMACPIPE,FMADD,e,b,d,e);
    make_inst(FMACPIPE,FNMSUB,f,b,c,f);
  }
  else {
    int t;

    make_inst(FMULPIPE,FMUL,e,a,c);

    make_inst(FMULPIPE,FMUL,f,a,d);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FSUB,f,f,t);
  }
}

void queue_two_cmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                     int e2,int f2,int a2,int b2,int c2,int d2)
{
  if ( have_fmadd ){
    queue_cmul(e1,f1,a1,b1,c1,d1);
    queue_cmul(e2,f2,a2,b2,c2,d2);
  }
  else{
    int t;

    make_inst(FMULPIPE,FMUL,e1,a1,c1);
    make_inst(FMULPIPE,FMUL,e2,a2,c2);
    make_inst(FMULPIPE,FMUL,f1,a1,d1);
    make_inst(FMULPIPE,FMUL,f2,a2,d2);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,d1);  
    make_inst(FADDPIPE,FSUB,e1,e1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,d2);
    make_inst(FADDPIPE,FSUB,e2,e2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,c1);
    make_inst(FADDPIPE,FADD,f1,f1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,c2);
    make_inst(FADDPIPE,FADD,f2,f2,t);

  }
  return;

}

void queue_two_cmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                      int e2,int f2,int a2,int b2,int c2,int d2)
{
  if ( have_fmadd ){
    queue_cmadd(e1,f1,a1,b1,c1,d1);
    queue_cmadd(e2,f2,a2,b2,c2,d2);
  }
  else{
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,a1,c1);
    make_inst(FADDPIPE,FADD,e1,e1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a2,c2);
    make_inst(FADDPIPE,FADD,e2,e2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,a1,d1);
    make_inst(FADDPIPE,FADD,f1,f1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,a2,d2);
    make_inst(FADDPIPE,FADD,f2,f2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,b1,d1);  
    make_inst(FADDPIPE,FSUB,e1,e1,t );

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,b2,d2);
    make_inst(FADDPIPE,FSUB,e2,e2,t );

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,b1,c1);
    make_inst(FADDPIPE,FADD,f1,f1,t );

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t ,b2,c2);
    make_inst(FADDPIPE,FADD,f2,f2,t );
  }
  return;

}



void queue_conjmadd(int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,e,a,c,e);
    make_inst(FMACPIPE,FMADD,f,a,d,f);
    make_inst(FMACPIPE,FMADD,e,b,d,e);
    make_inst(FMACPIPE,FNMSUB,f,b,c,f);
  }
  else {
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,f,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FADD,e,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FSUB,f,f,t);
  }
}

void queue_conjmaddto(int h, int g,int e,int f,int a,int b,int c,int d)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,h,a,c,e);
    make_inst(FMACPIPE,FMADD,g,a,d,f);
    make_inst(FMACPIPE,FMADD,h,b,d,h);
    make_inst(FMACPIPE,FNMSUB,g,b,c,g);
  }
  else {
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,c);
    make_inst(FADDPIPE,FADD,h,e,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,d);
    make_inst(FADDPIPE,FADD,g,f,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,d);
    make_inst(FADDPIPE,FADD,h,h,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b,c);
    make_inst(FADDPIPE,FSUB,g,g,t);
  }
}

void queue_two_conjmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                         int e2,int f2,int a2,int b2,int c2,int d2)
{
  if ( have_fmadd ){
    queue_conjmadd(e1,f1,a1,b1,c1,d1);
    queue_conjmadd(e2,f2,a2,b2,c2,d2);
  }
  else{
    int t;
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a1,c1);
    make_inst(FADDPIPE,FADD,e1,e1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a2,c2);
    make_inst(FADDPIPE,FADD,e2,e2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a1,d1);
    make_inst(FADDPIPE,FADD,f1,f1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a2,d2);
    make_inst(FADDPIPE,FADD,f2,f2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,d1);  
    make_inst(FADDPIPE,FADD,e1,e1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,d2);
    make_inst(FADDPIPE,FADD,e2,e2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,c1);
    make_inst(FADDPIPE,FSUB,f1,f1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,c2);
    make_inst(FADDPIPE,FSUB,f2,f2,t);
  }
  return;

}

void queue_two_conjmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                        int e2,int f2,int a2,int b2,int c2,int d2)
{
  if ( have_fmadd ){
    queue_conjmul(e1,f1,a1,b1,c1,d1);
    queue_conjmul(e2,f2,a2,b2,c2,d2);
  }
  else{
    int t;

    make_inst(FMULPIPE,FMUL,e1,a1,c1);
    make_inst(FMULPIPE,FMUL,e2,a2,c2);
    make_inst(FMULPIPE,FMUL,f1,a1,d1);
    make_inst(FMULPIPE,FMUL,f2,a2,d2);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,d1);  
    make_inst(FADDPIPE,FADD,e1,e1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,d2);
    make_inst(FADDPIPE,FADD,e2,e2,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b1,c1);
    make_inst(FADDPIPE,FSUB,f1,f1,t);

    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,b2,c2);
    make_inst(FADDPIPE,FSUB,f2,f2,t);
  }
  return;

}



void queue_three_cmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                       int e2,int f2,int a2,int b2,int c2,int d2,
                       int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ){
    make_inst(FMULPIPE,FMUL,e1,a1,c1);
    make_inst(FMULPIPE,FMUL,f1,a1,d1);

    make_inst(FMULPIPE,FMUL,e2,a2,c2);
    make_inst(FMULPIPE,FMUL,f2,a2,d2);

    make_inst(FMULPIPE,FMUL,e3,a3,c3);
    make_inst(FMULPIPE,FMUL,f3,a3,d3);

    make_inst(FMACPIPE,FNMSUB,e1,b1,d1,e1);
    make_inst(FMACPIPE,FMADD,f1,b1,c1,f1);

    make_inst(FMACPIPE,FNMSUB,e2,b2,d2,e2);
    make_inst(FMACPIPE,FMADD,f2,b2,c2,f2);

    make_inst(FMACPIPE,FNMSUB,e3,b3,d3,e3);
    make_inst(FMACPIPE,FMADD,f3,b3,c3,f3);
  }
  else{
    queue_cmul(e1,f1,a1,b1,c1,d1);
    queue_cmul(e2,f2,a2,b2,c2,d2);
    queue_cmul(e3,f3,a3,b3,c3,d3);
  }
  return;
}

void queue_three_cmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                        int e2,int f2,int a2,int b2,int c2,int d2,
                        int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,e1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,f1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,e2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,f2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,e3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,f3,a3,d3,f3);

    make_inst(FMACPIPE,FNMSUB,e1,b1,d1,e1);
    make_inst(FMACPIPE,FMADD,f1,b1,c1,f1);

    make_inst(FMACPIPE,FNMSUB,e2,b2,d2,e2);
    make_inst(FMACPIPE,FMADD,f2,b2,c2,f2);

    make_inst(FMACPIPE,FNMSUB,e3,b3,d3,e3);
    make_inst(FMACPIPE,FMADD,f3,b3,c3,f3);
  } 
  else {
    queue_cmadd(e1,f1,a1,b1,c1,d1);
    queue_cmadd(e2,f2,a2,b2,c2,d2);
    queue_cmadd(e3,f3,a3,b3,c3,d3);
  }
  return;
}



void queue_three_cmaddtos(
                       int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
                       int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
                       int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ) {
    make_inst(FMACPIPE,FMADD,h1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,g1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,h2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,g2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,h3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,g3,a3,d3,f3);

    make_inst(FMACPIPE,FNMSUB,h1,b1,d1,h1);
    make_inst(FMACPIPE,FMADD,g1,b1,c1,g1);

    make_inst(FMACPIPE,FNMSUB,h2,b2,d2,h2);
    make_inst(FMACPIPE,FMADD,g2,b2,c2,g2);

    make_inst(FMACPIPE,FNMSUB,h3,b3,d3,h3);
    make_inst(FMACPIPE,FMADD,g3,b3,c3,g3);
  } 
  else {
    queue_cmaddto(h1,g1,e1,f1,a1,b1,c1,d1);
    queue_cmaddto(h2,g2,e2,f2,a2,b2,c2,d2);
    queue_cmaddto(h3,g3,e3,f3,a3,b3,c3,d3);
  }
  return;
}


void queue_three_conjmadds(int e1,int f1,int a1,int b1,int c1,int d1,
                           int e2,int f2,int a2,int b2,int c2,int d2,
                           int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ){
    make_inst(FMACPIPE,FMADD,e1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,f1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,e2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,f2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,e3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,f3,a3,d3,f3);

    make_inst(FMACPIPE,FMADD,e1,b1,d1,e1);
    make_inst(FMACPIPE,FNMSUB,f1,b1,c1,f1);

    make_inst(FMACPIPE,FMADD,e2,b2,d2,e2);
    make_inst(FMACPIPE,FNMSUB,f2,b2,c2,f2);

    make_inst(FMACPIPE,FMADD,e3,b3,d3,e3);
    make_inst(FMACPIPE,FNMSUB,f3,b3,c3,f3);
  }
  else{
    queue_conjmadd(e1,f1,a1,b1,c1,d1);
    queue_conjmadd(e2,f2,a2,b2,c2,d2);
    queue_conjmadd(e3,f3,a3,b3,c3,d3);
  }
  return;

}



void queue_three_conjmaddtos(
                       int h1,int g1,int e1,int f1,int a1,int b1,int c1,int d1,
                       int h2,int g2,int e2,int f2,int a2,int b2,int c2,int d2,
                       int h3,int g3,int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ){

    make_inst(FMACPIPE,FMADD,h1,a1,c1,e1);
    make_inst(FMACPIPE,FMADD,g1,a1,d1,f1);

    make_inst(FMACPIPE,FMADD,h2,a2,c2,e2);
    make_inst(FMACPIPE,FMADD,g2,a2,d2,f2);

    make_inst(FMACPIPE,FMADD,h3,a3,c3,e3);
    make_inst(FMACPIPE,FMADD,g3,a3,d3,f3);

    make_inst(FMACPIPE,FMADD,h1,b1,d1,h1);
    make_inst(FMACPIPE,FNMSUB,g1,b1,c1,g1);

    make_inst(FMACPIPE,FMADD,h2,b2,d2,h2);
    make_inst(FMACPIPE,FNMSUB,g2,b2,c2,g2);

    make_inst(FMACPIPE,FMADD,h3,b3,d3,h3);
    make_inst(FMACPIPE,FNMSUB,g3,b3,c3,g3);

  }
  else{

    queue_conjmaddto(h1,g1,e1,f1,a1,b1,c1,d1);
    queue_conjmaddto(h2,g2,e2,f2,a2,b2,c2,d2);
    queue_conjmaddto(h3,g3,e3,f3,a3,b3,c3,d3);

  }
  return;
}


void queue_three_conjmuls(int e1,int f1,int a1,int b1,int c1,int d1,
                          int e2,int f2,int a2,int b2,int c2,int d2,
                          int e3,int f3,int a3,int b3,int c3,int d3)
{
  if ( have_fmadd ){
    make_inst(FMULPIPE,FMUL,e1,a1,c1);
    make_inst(FMULPIPE,FMUL,f1,a1,d1);

    make_inst(FMULPIPE,FMUL,e2,a2,c2);
    make_inst(FMULPIPE,FMUL,f2,a2,d2);

    make_inst(FMULPIPE,FMUL,e3,a3,c3);
    make_inst(FMULPIPE,FMUL,f3,a3,d3);

    make_inst(FMACPIPE,FMADD,e1,b1,d1,e1);
    make_inst(FMACPIPE,FNMSUB,f1,b1,c1,f1);

    make_inst(FMACPIPE,FMADD,e2,b2,d2,e2);
    make_inst(FMACPIPE,FNMSUB,f2,b2,c2,f2);

    make_inst(FMACPIPE,FMADD,e3,b3,d3,e3);
    make_inst(FMACPIPE,FNMSUB,f3,b3,c3,f3);
  }
  else{
    queue_conjmul(e1,f1,a1,b1,c1,d1);
    queue_conjmul(e2,f2,a2,b2,c2,d2);
    queue_conjmul(e3,f3,a3,b3,c3,d3);
  }
  return;

}
void queue_fmac ( int d, int a , int b ) 
{
  queue_fmadd(d,a,b,d);
}
void queue_fnmac ( int d, int a , int b ) 
{
  queue_fnmsub(d,a,b,d);
}

void queue_fnmsub ( int d, int a , int b, int c) 
{
  int t;
  if ( have_fmadd ){
    make_inst(FMACPIPE,FNMSUB,d,a,b,c);
  }
  else {
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,b);
    make_inst(FADDPIPE,FSUB,d,c,t);
  }
}
void queue_fmadd ( int d, int a , int b, int c ) 
{
  int t;
  if ( have_fmadd ){
    make_inst(FMACPIPE,FMADD,d,a,b,c);
  }
  else {
    t = get_rotating_register(CMADregs);
    make_inst(FMULPIPE,FMUL,t,a,b);
    make_inst(FADDPIPE,FADD,d,c,t);
  }
}
/* ...END complex queueing routines*/



