/*
 *
 *  Copyright Peter Boyle and Glasgow University 2000.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include "processor.h"

extern struct processor *PROC;
void do_writehint ( int ptr,int line );

/* {...Cache Streaming related functions*/
#define MAX_PLINES 24
#define MAX_LOOKUP 4
#define MAX_MULTIWAY 5
int lookupno =0;
char *gnames[MAX_LOOKUP]={"LOOKUP0","LOOKUP1","LOOKUP2","LOOKUP3"};

char *pnames[MAX_PLINES]={"PRE0","PRE1","PRE2","PRE3","PRE4",
                          "PRE5","PRE6","PRE7","PRE8","PRE9",
                          "PRE10","PRE11","PRE12","PRE13","PRE14",
                          "PRE15","PRE16","PRE17","PRE18","PRE19",
                          "PRE20","PRE21","PRE22","PRE23"
                         };
char *prnames[MAX_PLINES]={"PRER0","PRER1","PRER2","PRER3","PRER4",
                           "PRER5","PRER6","PRER7","PRER8","PRER9",
                           "PRER10","PRER11","PRER12","PRER13","PRER14",
                           "PRER15","PRER16","PRER17","PRER18","PRER19",
                           "PRER20","PRER21","PRER22","PRER23"
                         };
int pregs[MAX_PLINES]={-1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1
                      };
int pimms[MAX_PLINES]={-1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1,-1,
                       -1,-1,-1,-1
                      };
int Prefetch;
int cmovreg=-1;
int scratch=-1;
int ahead_imm=-1;
int shift_ahead_imm=-1;
int lookupmulti_imms[MAX_MULTIWAY+1]= { 
  -1,-1,-1,-1,-1,-1
};

const char *lookupmulti_immstr[MAX_MULTIWAY+1]= { 
  "zeromulti","onemulti","twomulti","threemulti","fourmulti","fivemulti"
};

int Bdy = -1; /*Immediate number for iterations check        */
              /*Switch to prefetching from stack when within */
              /*Bdy sites of inner loop                      */

int Isize = -1;
int minusone = -1;

void need_cache_line(int line);
int need_pref_regs ( void );
void do_prefetch ( int ptr,int line );
void set_lookup_pointer (struct stream *mystream);

void initialise_streams (void )
{
  int i;
  char *str;
  test_proc();

  Prefetch   = def_byte_offset( PROC->CacheLine,"CACHELINE");
  Isize      = def_byte_offset(PROC->I_size,"Isize");
  Bdy        = def_byte_offset(-(N_Ahead+2),"Bdy");
  ahead_imm  = def_byte_offset(N_Ahead,"Nahead");
  shift_ahead_imm  = def_byte_offset(N_Ahead*PROC->I_size,"SHFT_PREF_IMM");

  for(i=0;i<=MAX_MULTIWAY;i++){
   lookupmulti_imms[i]=
        def_byte_offset(i*PROC->I_size,lookupmulti_immstr[i]);
  }

  cmovreg = allocate_reg(Iregs,"predicate");
  scratch = allocate_reg(Iregs,"prefscratch");
  minusone = def_byte_offset(-1,"minusone");
}

struct stream * create_stream(int atom,int base_reg,int countreg,
                               int dir, int kind,...)
{
  struct stream *ret = malloc ( sizeof(struct stream));
  int size = get_offset(atom);
  int i,line;
  va_list ap;

  va_start ( ap,kind);

  ret->inout = dir;
  ret->pointer = base_reg;
  ret->basereg = base_reg;
  ret->atom = atom;
  ret->counter = countreg;
  ret->pattern = kind;
        /*no of cache Lines covering one atom*/
  ret->lines = 1  + ((size-1 )/(PROC->CacheLine)); 

  if ( dir != STREAM_IN && dir != STREAM_OUT ) {
    puts("Create stream: direction not specified");
    exit(1);
  }

  switch ( kind ){
  case LINEAR:
    ret->stride = ret->atom;
    ret->lookupreg = -1;
    ret->minline = (N_Ahead*size)/PROC->CacheLine;
    /*Check we have grabbed the offsets required   */
    /*On powerpc these have to be stored in Iregs  */
    /*since Iops are at a premium and the dcbt inst*/
    /*doesn't have and immediate offset            */
    for ( i = 0 ; i < ret->lines ; i++ ){
      line =  i + (N_Ahead*size)/PROC->CacheLine ;
      need_cache_line(line);
    }
    break;
  case STRIDED:
    ret->stride = va_arg(ap,int); /*Stride should be an offset handle, */
                                  /*not the number itself*/
    ret->lookupreg = -1;
    ret->minline = 0;
    for ( i = 0 ; i < ret->lines ; i++ ){
      need_cache_line(i);
    }
    break;
  case POINTER:
    ret->basereg = -1;
    ret->stride = -1;
    ret->lookupreg = va_arg(ap,int); /*This is the table pointer*/
    set_lookup_pointer(ret);
    break;
  case LOOKUP:
    /*We need to keep a copy of the start of the array*/
    if ( lookupno >= MAX_LOOKUP ) { 
       puts("Error out of indirection regs");
       exit(0);
    }
    ret->basereg = allocate_reg(Iregs,gnames[lookupno++]);
    make_inst(IALUPIPE,IOR,ret->basereg,base_reg,base_reg);
    ret->stride = -1;
    ret->lookupreg = va_arg(ap,int);
    ret->minline = 0;
    for ( i = 0 ; i < ret->lines ; i++ ){
      need_cache_line(i);
    }
    need_cache_line(1);
    set_lookup_pointer(ret);
    break;

  case LOOKUP_MULTI: /*A multi way look up table*/

    /*We need to keep a copy of the start of the array*/
    if ( lookupno >= MAX_LOOKUP ) { 
       puts("Error out of indirection regs");
       exit(0);
    }
    ret->basereg = allocate_reg(Iregs,gnames[lookupno++]);
    make_inst(IALUPIPE,IOR,ret->basereg,base_reg,base_reg);
    ret->stride = -1;
    ret->way  = va_arg(ap,int); /*Lookup field way in */
    ret->lookupreg = va_arg(ap,int);
    ret->nway = va_arg(ap,int); /*a table with nway ints per site*/
    ret->minline = 0;
    for ( i = 0 ; i < ret->lines ; i++ ){
      need_cache_line(i);
    }
    need_cache_line(1);
    set_lookup_pointer(ret);
    break;

  default:
    puts("create_stream: Unkown access pattern");
    exit(1);
  }
  return(ret);
}

void need_cache_line(int line)
{
  if ( line >= MAX_PLINES ) { printf("MAX_PLINES too small for %d\n",line); exit(0);}

  if ( pimms[line] == -1 ) {
       pimms[line] = def_byte_offset(line*PROC->CacheLine,pnames[line]);
  }

  /*Power doesn't provide immediate offsets for prefetches   */
  /*So stick the value in a register to save recalculation of*/
  /*offset value                                             */
  if ( need_pref_regs() ){
    if ( pregs[line]== -1 ){
      pregs[line] = allocate_reg(Iregs,prnames[line]);
      make_inst(LOADPIPE,LOAD_IMM,pregs[line],pimms[line]);
    }
  }
  return;
}

int need_pref_regs ( void )
{
  return( PROC->iset->instructions[CACHPIPE][PREF_IMM].argtypes[0] == Iregs );
}

void do_prefetch ( int ptr,int line )
{
      if ( need_pref_regs() ){  
        if (pregs[line]<0) {puts("Error: unallocated pref line"); exit(-1);}
        make_inst(CACHPIPE,PREF_IMM,pregs[line],ptr);
      } 
      else {
        make_inst(CACHPIPE,PREF_IMM,pimms[line],ptr);
      }
}
void do_flush( int ptr, int line ) 
{
      if ( need_pref_regs() ){  
        if (pregs[line]<0) {puts("Error: unallocated pref line"); exit(-1);}
        make_inst(CACHPIPE,FLUSH,pregs[line],ptr);
      } 
      else {
	/* Not supported on anything other than PPC yet */
	/*       make_inst(CACHPIPE,PREF_IMM,pimms[line],ptr); */
	return;
      }
}
void do_writehint ( int ptr,int line )
{
  /*
   *
   */
    if ( need_pref_regs() ){  
      if (pregs[line]<0) {puts("Error: unallocated pref line"); exit(-1);}
      make_inst(CACHPIPE,WRITE_HINT,pregs[line],ptr);
      if ( pregs[line] == 0 ) { 
	puts("do_writehint: Error reg zero is weird");
	exit(0);
      }
    } 
    else {
      queue_load_addr(scratch,pimms[line],ptr);
      make_inst(CACHPIPE,WRITE_HINT,scratch);
    }
}

void queue_prefetch  (struct stream *mystream)
{
  int i;
  int line;

  switch ( mystream->pattern ){
  case LINEAR:
    for(i=0;i<mystream->lines;i++){
      line = mystream->minline + i; 
      do_prefetch(mystream->basereg,line);
    }
    break;
  case STRIDED:
    queue_iload_imm(scratch,mystream->stride);
    queue_imul_imm(scratch,scratch,ahead_imm);
    queue_iadd(scratch,scratch,mystream->basereg);
    for(i=0;i<mystream->lines;i++){
      do_prefetch(scratch,i);
    }
    break;
  case LOOKUP:
    /*Load 3 ahead in shift table => scratch */  
    queue_iload(scratch,shift_ahead_imm,mystream->lookupreg);
    /*  times atom to get byte offset into array*/
    queue_imul_imm(scratch,scratch,mystream->atom);
    /* add to base register */
    queue_iadd(scratch,mystream->basereg,scratch);
    /* Cmov to stack pointer if off limits in array*/
    queue_bounds_check ( mystream->counter,scratch);
    
    for(i=0;i<mystream->lines;i++){
      do_prefetch(scratch,i);
    }
    do_prefetch(mystream->lookupreg,1);
    break;
  case LOOKUP_MULTI:
    set_lookup_nxt_pointer(mystream,scratch);    
    for(i=0;i<mystream->lines;i++){
      do_prefetch(scratch,i);
    }
  }
}

void set_lookup_pointer (struct stream *mystream)
{
  if ( mystream -> pattern == LOOKUP ){

    queue_iload(scratch,pimms[0],mystream->lookupreg);
    queue_imul_imm(scratch,scratch,mystream->atom);
    /* add to base register and store in pointer*/
    queue_iadd(mystream->pointer,mystream->basereg,scratch);

  }  else if ( mystream-> pattern == LOOKUP_MULTI ){

   /*Load our way in shift table => scratch */  
   queue_iload(scratch,lookupmulti_imms[mystream->way],mystream->lookupreg);

  }  else if ( mystream-> pattern == POINTER ){

    queue_iload(mystream->pointer,pimms[0],mystream->lookupreg);

  } else {

    puts("Error: set_lookup_basereg on non-lookup stream");
    exit(0);

  }

}



void set_lookup_nxt_pointer (struct stream *mystream,int nxt)
{


  if ( mystream -> pattern == LOOKUP ){

    queue_iload(scratch,Isize,mystream->lookupreg);
    queue_imul_imm(scratch,scratch,mystream->atom);
    queue_iadd(nxt,mystream->basereg,scratch);
    queue_load_addr(scratch,minusone,mystream->counter);
    queue_cmovle(scratch,PROC->StackPointer,nxt);

  } else if ( mystream -> pattern == LOOKUP_MULTI ){

    queue_load_addr(scratch,lookupmulti_imms[mystream->nway],mystream->lookupreg);
    queue_iload(nxt,lookupmulti_imms[mystream->way],scratch);
    queue_load_addr(scratch,minusone,mystream->counter);
    queue_cmovle(scratch,PROC->StackPointer,nxt);

  } else {
    puts("Error: set_lookup_basereg on non-lookup stream");
    exit(0);
  }

}

void iterate_table(struct stream * mystream)
{
  if ( mystream -> pattern == LOOKUP_MULTI ){
    queue_load_addr(mystream->lookupreg,
        lookupmulti_imms[mystream->nway],mystream->lookupreg);
    do_prefetch(mystream->lookupreg,1);
  } else {
    puts("iterate_table: pooh"); exit(0);
  }
}

void iterate_stream(struct stream * mystream)
{

  if ( mystream -> pattern == LOOKUP ){
    queue_load_addr(mystream->lookupreg,Isize,mystream->lookupreg);
    set_lookup_pointer(mystream);
  } else if ( mystream -> pattern == POINTER ) {
    queue_load_addr(mystream->lookupreg,Isize,mystream->lookupreg);
    set_lookup_pointer(mystream);
  } else if ( mystream -> pattern == LOOKUP_MULTI ) {
    set_lookup_pointer(mystream);
  } else if ( mystream -> pattern == STRIDED ){
    queue_load_addr(mystream->basereg,mystream->stride,mystream->basereg);
  } else {
    queue_load_addr(mystream->basereg,mystream->atom,mystream->basereg);
  }
}

void queue_pref_set_predicate( int counter )
{
  queue_load_addr(cmovreg,Bdy,counter);
  if ( PROC->iset->instructions[IALUPIPE][IOR_PREDICATE].mnemonic !=NULL ){
    make_inst(IALUPIPE,IOR_PREDICATE,cmovreg,cmovreg,cmovreg);
  }
}

void queue_bounds_check ( int counter, int reg )
{
  queue_cmovlt(cmovreg,PROC->StackPointer,reg);
}

void queue_write_hint(struct stream *mystream)
{
  int i;

  if ( mystream ->pattern == LINEAR ){
    queue_load_addr(scratch,pimms[mystream->minline],mystream->basereg);
    queue_bounds_check(mystream->counter,scratch);

    for(i=0;i<mystream->lines;i++){
      make_inst(CACHPIPE,WRITE_HINT,scratch);
      if ( i+1 <mystream->lines )
        queue_load_addr(scratch,Prefetch,scratch);
    }
  }
  else {
    queue_prefetch(mystream);
  }
}
